#!%PERL%
# Copyright (c) vhffs project and its contributors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 3. Neither the name of vhffs nor the names of its contributors
#    may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# 

use strict;
use utf8;

use lib '%VHFFS_LIB_DIR%';
use Term::ReadPassword;
use File::Basename;
use File::Temp;
use DBI;
use Cwd qw(abs_path getcwd);
use Vhffs::Main;
use IO::File;

#my $SQL_DIR = '%VHFFS_BACKEND_DIR%/';
my $SQL_DIR = '/root/vhffs/vhffs/trunk/vhffs-backend/src/pgsql/';


sub confirm($) {
    my $text = shift;
    my $answer;
    do {
        print "$text (Y/n) ? ";
        chomp($answer = <STDIN>);
        $answer = lc($answer);
    } while($answer ne 'y' and $answer ne 'n' and $answer ne '');

    return($answer ne 'n');
}

sub upgrade_40_pre1() {
    my $script = '%VHFFS_COMPAT_DIR%/4.0pre1.sql';
    system("psql -f $script");
}

sub upgrade_40() {
    my $script = '%VHFFS_COMPAT_DIR%/4.0.sql';
    system("psql -f $script");
}

sub is_40_pre1($) {
    # seq_id_groups only exists in pre1, it isn't created in 4.0
    my $dbh = shift;
    my ($res) = $dbh->selectrow_array('SELECT count(*) FROM pg_catalog.pg_class WHERE relname=\'seq_id_groups\'');
    return $res;
}

sub is_40($) {
    my $dbh = shift;
    my ($res) = $dbh->selectrow_array('select count(*) from pg_catalog.pg_constraint where conname = \'vhffs_users_unique_username\'');
    return ($res == 0);

}

print "**** VHFFS UPDATEDB ****\nThis script will upgrade your database to the VHFFS 4.1's schema.\nEnsure that VHFFS is *stopped* !\nUse this script to your own risks!!!\n";
if(!confirm('Do you still want to continue')) {
    exit(-1);
}

my ($dbhost, $dbname, $dbuser, $dbpass);

print 'Enter VHFFS DB hostname [localhost] : ';
chomp($dbhost = <STDIN>);
$dbhost = 'localhost' unless($dbhost);
print 'Enter VHFFS DB name [vhffs] : ';
chomp($dbname = <STDIN>);
$dbname = 'vhffs' unless($dbname);
print 'Enter VHFFS DB username [vhffs] : ';
chomp($dbuser = <STDIN>);
$dbuser = 'vhffs' unless($dbuser);
$dbpass = read_password('Enter VHFFS DB password : ');

my $dbh = DBI->connect("DBI:Pg:dbname=$dbname;host=$dbhost;port=5432",$dbuser, $dbpass);
if(!$dbh) {
    die "Cant connect to VHFFS DB\n";
}

`psql --version` =~ /^.*?(\d+)/;
my $version = $1;
if($version == 7) {
    print "You may be asked for you pgsql password during upgrade (psql 7.x doesn't support file authentication).\n";
}

# psql and pg_dump will not ask for a password using this
my $pgpass = new File::Temp(DIR => '/tmp');
print $pgpass "*:*:$dbname:$dbuser:$dbpass"; 
$ENV{'PGHOST'} = $dbhost;
$ENV{'PGDATABASE'} = $dbname;
$ENV{'PGUSER'} = $dbuser;
$ENV{'PGPASSFILE'} = $pgpass->filename;

if(confirm('Do you want to perform a backup right now')) {
    print 'Enter a filename for the backup file [/tmp/vhffs-backup.sql] : ';
    my $backupfile = <STDIN>;
    chomp($backupfile);
    $backupfile = '/tmp/vhffs-backup.sql' unless($backupfile);
    my $bfh = new IO::File($backupfile, 'w');
    die("Unable to open $backupfile : $!\n") unless($bfh);
    my $pg_dump = new IO::File("pg_dump |");
    while(<$pg_dump>) {
        print $bfh $_;
    }
    $bfh->close();
}

if(!is_40($dbh) && !confirm('It seems that you are not running vhffs 4.0, do you still want to upgrade')) {
    exit(-2);
} else {
    if(is_40_pre1($dbh)) {
        print "Detected VHFFS 4.0-pre1, performing transition upgrade\n";
        upgrade_40_pre1();
    }
}

upgrade_40();

my $INIT_DB = new IO::File($SQL_DIR.'initdb.sql', 'r');
die("unable to open ${SQL_DIR}.initdb.sql : $!, DB not upgraded\n") unless($INIT_DB);

print "Let's drop foreign keys...\n";
flush STDOUT;

# Fetches and removes all foreign key constraints

my $fk = $dbh->selectall_arrayref('SELECT c.conname AS constraint, cl.relname AS table FROM pg_constraint c INNER JOIN pg_class cl ON c.conrelid = cl.oid WHERE c.contype = \'f\' AND cl.relname LIKE \'vhffs_%\'', { Slice => {} });

foreach(@$fk) {
    print "\tDropping $_->{table}($_->{constraint})\n";
    $dbh->do("ALTER TABLE $_->{table} DROP CONSTRAINT \"$_->{constraint}\" CASCADE") or warn "Unable to drop constraint $_->{constraint}\n";
}

print " done.\nLet's drop unique key constraints\n";

my $uc = $dbh->selectall_arrayref('SELECT c.conname AS constraint, cl.relname AS table FROM pg_catalog.pg_constraint c INNER JOIN pg_class cl ON c.conrelid = cl.oid WHERE contype = \'u\' AND cl.relname LIKE \'vhffs_%\'', { Slice => {} });

foreach(@$uc) {
    print "\tDropping $_->{table}($_->{constraint})\n";
    $dbh->do("ALTER TABLE $_->{table} DROP CONSTRAINT \"$_->{constraint}\" CASCADE") or warn "Unable to drop constraint $_->{constraint}\n";
}

print " done.\nLet's drop indexes\n";

my $idx = $dbh->selectall_arrayref('SELECT c.relname AS index FROM pg_catalog.pg_class c INNER JOIN pg_catalog.pg_index i ON i.indexrelid = c.oid INNER JOIN pg_catalog.pg_class c2 ON c2.oid = i.indrelid WHERE i.indisprimary = \'f\' AND i.indisunique = \'f\' AND c2.relname LIKE \'vhffs_%\'', { Slice => {} });

foreach(@$idx) {
    print "\tDropping index $_->{index}\n";
    $dbh->do("DROP INDEX \"$_->{index}\"") or warn "Unable to drop index $_->{index}\n";
}

print " done.\n";


# Create new foreign key constraints based on initdb.sql content

print "Creating new foreign key, indexes and unique keys...\n";
flush STDOUT;

while(<$INIT_DB>) {
    if(/(ALTER TABLE\s+(.*?)\s+ADD CONSTRAINT\s+(.*?)\s+FOREIGN KEY.*);\s*$/i) {
        print "\tCreating $2($3)\n";
        $dbh->do($1) or warn("*** Unable to create constraint $3 ***\n");
    } elsif(/(ALTER TABLE\s+(.*?)\s+ADD CONSTRAINT\s+(.*?)\s+UNIQUE.*);\s*$/i) {
        print "\tCreating $2($3)\n";
        $dbh->do($1) or warn("*** Unable to create constraint $3 ***\n");
    } elsif(/(CREATE INDEX\s+(.*?)\s+ON\s(.*));\s*$/i) {
        print "\tCreating $2 ON $3\n";
        $dbh->do($1) or warn("*** Unable to create index $3 ***\n");
    }
}

$INIT_DB->close();

print "Fixing \\s in objects' description...";
flush STDOUT;

my $rows = $dbh->selectall_arrayref('SELECT object_id, description FROM vhffs_object');

my $sth = $dbh->prepare('UPDATE vhffs_object SET description = ? WHERE object_id = ?');
foreach(@{$rows}) {
    $_->[1] =~ s/\\([^\\])/$1/g;
    $sth->execute($_->[1], $_->[0]) or warn "Unable to update object #$_->[0] (description : $_->[1])\n";
}

print " done.\n";


print "Fixing \\s in objects' history...";
flush STDOUT;

$rows = $dbh->selectall_arrayref('SELECT history_id, message FROM vhffs_history');

$sth = $dbh->prepare('UPDATE vhffs_history SET message = ? WHERE history_id = ?');
foreach(@$rows) {
    $_->[1] =~ s/\\([^\\])/$1/g;
    $sth->execute($_->[1], $_->[0]) or warn "Unable to update history #$_->[0] (message: $_->[1])\n";
}

print " done.\n";

print "Fixing \\s in users' informations...";
flush STDOUT;

$rows = $dbh->selectall_arrayref('SELECT uid, shell FROM vhffs_users');

$sth = $dbh->prepare('UPDATE vhffs_users SET shell = ? WHERE uid = ?');
foreach(@$rows) {
    $_->[1] =~ s/\\([^\\])/$1/g;
    $sth->execute($_->[1], $_->[0]) or warn "Unable to update user #$_->[0] (shell : $_->[1])\n";
}

$rows = $dbh->selectall_arrayref('SELECT uid, firstname, lastname, address, city, country FROM vhffs_users', { Slice => {} });

$sth = $dbh->prepare('UPDATE vhffs_users SET firstname = ?, lastname = ?, address = ?, city = ?, country = ? WHERE uid = ?');
foreach(@$rows) {
    foreach my $prop (qw(firstname lastname address, city, country) ) {
        $_->{$prop} =~ s/\\([^\\])/$1/g;
    }
    $sth->execute($_->{firstname}, $_->{lastname}, $_->{address}, $_->{city}, $_->{country}, $_->{uid}) or warn "Unable to update infos for user #$_{uid}";
}

print " done.\n";
